#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_simtool
----------------------------------

Tests for `simtool` module.
"""

import pytest
import simtool


@pytest.fixture
def response():
    """Sample pytest fixture.
    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument.
    """
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string



# @package      hubzero-simtool
# @file         utils.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import sys
import re
import glob
import nbformat
from papermill.iorw import load_notebook_node
import yaml
import jsonpickle
from .params import Params

def parse(inputs):
   """Convert YAML expression of SimTool input or outputs into a collection
      of Params objects

      Args:
          inputs: YAML expression of SimTool inputs or outputs

      Returns:
          d: dictionary of Params objects.  Each Params objects represents
              one SimTool input or output.
   """
   d = Params()
   for i in inputs:
      t = inputs[i]['type']
      if t in Params.types:
         d[i] = Params.types[t](**inputs[i])
      else:
         print('Unknown type:', t, file=sys.stderr)
   return d


def getValidatedInputs(inputs):
   """Test inputs for validity.

      Any invalid inputs are reported.  This function is intended for use
      in the 'parameters' cell.  The following code will set parameters to
      default values and cause an error if default values are in error.

      Note:
          from simtool import getValidatedInputs

          defaultInputs = getValidatedInputs(INPUTS)
          if defaultInputs: globals().update(defaultInputs)

      Args:
          inputs: YAML representation of SimTool inputs.

      Returns:
          dictionary of Params.
   """
   validatedInputs = {}
   try:
      params = parse(inputs)
   except ValueError as e:
      print(e)
   else:
      for param in params:
         validatedInputs[param] = params[param].value

   return validatedInputs


def _get_extra_files(nbPath):
   """Internal function to search the notebook for a cell tagged
   'FILES' with content 'EXTRA_FILES=xxx' where 'xxx' is a list of files
   or '*'
   """
   ecell = None
   nb = load_notebook_node(nbPath)
   for cell in nb.cells:
      if 'FILES' in cell.metadata.tags:
         ecell = cell['source']
         break
   if ecell is None:
      return None

   extra = None
   for line in ecell.split('\n'):
      if line.startswith('EXTRA_FILES'):
         extra = line
         break
   if extra is None:
      print("WARNING: cannot parse FILE cell:")
      return None

   try:
      val = extra.split('=')[1].replace("'", '"')
      return jsonpickle.loads(val)
   except:
      print("WARNING: cannot parse:", extra)

   return None


def _getSimToolDescription(nbPath):
   """Internal function to search the notebook for a cell tagged
   'DESCRIPTION' with content 'DESCRIPTION=xxx' where 'xxx' is a
   string describing the simtool
   """
   ecell = None
   nb = load_notebook_node(nbPath)
   for cell in nb.cells:
      if 'DESCRIPTION' in cell.metadata.tags:
         ecell = cell['source']
         break
   if ecell is None:
      return None

   descriptionText = ecell.split('=')[1:]
   descriptionText = ''.join(descriptionText)
   descriptionText = descriptionText.strip(' ')
   if   descriptionText.startswith('"""') and descriptionText.endswith('"""'):
      descriptionText = descriptionText[3:-3]
   elif descriptionText.startswith('"') and descriptionText.endswith('"'):
      descriptionText = descriptionText[1:-1]
   elif descriptionText.startswith("'") and descriptionText.endswith("'"):
      descriptionText = descriptionText[1:-1]

   return descriptionText


def getGetSimToolNameRevisionFromEnvironment():
   """Determine the SimTool name and revision from environment set by submit

      Returns:
          simToolName: SimTool name set by submit
          simToolRevision: SimTool revision set by submit
   """
   simToolName     = None
   simToolRevision = None
   try:
      submitApplicationRevision = os.environ['SUBMIT_APPLICATION_REVISION']
   except:
      pass
   else:
      varParts = submitApplicationRevision.split('_')
      simToolRevision = varParts[-1]
      simToolName     = '_'.join(varParts[0:-1])

   return simToolName,simToolRevision


def _getSimToolNotebookMetaData(nbPath):
   simToolNotebookMetaData = {}
   simToolNotebookMetaData['name']     = None
   simToolNotebookMetaData['revision'] = None
   simToolNotebookMetaData['state']    = None

   try:
      nb = nbformat.read(nbPath,nbformat.NO_CONVERT)
   except:
      pass
   else:
      try:
         metadata = nb['metadata']['simTool_info']
      except (AttributeError,KeyError) as err:
         pass
      else:
         try:
            name = metadata['name']
         except:
            pass
         else:
            simToolNotebookMetaData['name'] = name

         try:
            revision = metadata['revision']
         except:
            pass
         else:
            simToolNotebookMetaData['revision'] = "r%d" % (revision)

         try:
            state = metadata['state']
         except:
            pass
         else:
            simToolNotebookMetaData['state'] = state

   return simToolNotebookMetaData


def findSimToolNotebook(simToolName,simToolRevision=None):
   """Lookup simtool by name and revision.

      This function has been replaced by searchForSimTool(simToolName,simToolRevision=None)
   """
   simToolLocation = {}
   simToolLocation['notebookPath']    = None
   simToolLocation['simToolName']     = None
   simToolLocation['simToolRevision'] = None
   simToolLocation['published']       = None

   if   simToolRevision and not simToolName.endswith('.ipynb'):
      simToolNotebook = os.path.basename(simToolName) + '.ipynb'
      notebookPath = os.path.join(os.sep,'apps',simToolName,simToolRevision,'simtool',simToolNotebook)
      if os.path.exists(notebookPath):
         # look for installed or published revision in /apps/name/revision/simtool/
         simToolLocation['notebookPath']    = os.path.realpath(notebookPath)
         simToolLocation['simToolName']     = os.path.basename(simToolName)
         simToolLocation['simToolRevision'] = os.path.basename(os.path.dirname(os.path.dirname(simToolLocation['notebookPath'])))
         # verify pubication status - sample published notebook reference to simtool
         simToolNotebookMetaData = _getSimToolNotebookMetaData(simToolLocation['notebookPath'])
         if simToolNotebookMetaData['name'] == simToolLocation['simToolName'] and \
            simToolNotebookMetaData['revision'] == simToolLocation['simToolRevision'] and \
            simToolNotebookMetaData['state'] == 'published':
            simToolLocation['published'] = True
         else:
            simToolLocation['published'] = False
#           print(simToolNotebookMetaData['name'],simToolLocation['simToolName'])
#           print(simToolNotebookMetaData['revision'],simToolLocation['simToolRevision'])
#           print(simToolNotebookMetaData['state'])
      else:
         notebookPath = os.path.join(simToolName,simToolRevision,'simtool',simToolNotebook)
         if os.path.exists(notebookPath):
            # look for notebook in name/revision/simtool/
            simToolLocation['notebookPath']    = os.path.realpath(notebookPath)
            simToolLocation['simToolName']     = os.path.basename(simToolName)
            simToolLocation['simToolRevision'] = simToolRevision
            simToolLocation['published']       = False
   elif not simToolName.endswith('.ipynb'):
      # revision not specified
      # look for latest published revision in /apps
      simToolNotebook = os.path.basename(simToolName) + '.ipynb'
      notebookPathPattern = os.path.join(os.sep,'apps',simToolName,'*','simtool',simToolNotebook)
      newestRevision = 0
      for notebookPath in glob.glob(notebookPathPattern):
         revision = notebookPath.split(os.sep)[3]
         if revision.startswith('r'):
            simToolNotebookMetaData = _getSimToolNotebookMetaData(notebookPath)
            if simToolNotebookMetaData['state'] == 'published':
               revisionNumber = int(revision[1:])
               if revisionNumber > newestRevision:
                  newestRevision = revisionNumber
                  simToolLocation['notebookPath']    = os.path.realpath(notebookPath)
                  simToolLocation['simToolName']     = os.path.basename(simToolName)
                  simToolLocation['simToolRevision'] = os.path.basename(os.path.dirname(os.path.dirname(simToolLocation['notebookPath'])))
                  simToolLocation['published']       = True

      if simToolLocation['published'] is None:
         notebookPath = os.path.join(simToolName,simToolNotebook)
         if os.path.exists(notebookPath):
            # look for notebook in name
            simToolLocation['notebookPath']    = os.path.realpath(notebookPath)
            simToolLocation['simToolName']     = os.path.basename(simToolName)
            simToolLocation['simToolRevision'] = None
            simToolLocation['published']       = False
         else:
            notebookPath = os.path.join(simToolName,'simtool',simToolNotebook)
            if os.path.exists(notebookPath):
               # look for notebook in name
               simToolLocation['notebookPath']    = os.path.realpath(notebookPath)
               simToolLocation['simToolName']     = os.path.basename(simToolName)
               simToolLocation['simToolRevision'] = None
               simToolLocation['published']       = False
               notebookPath = simToolLocation['notebookPath'].split(os.sep)
               notebookPath.pop(0)
               if len(notebookPath) == 5:
                  revision = notebookPath.pop(2)
                  if os.sep.join(notebookPath) == \
                     os.path.join('apps',simToolLocation['simToolName'],'simtool',simToolNotebook):
                     simToolLocation['simToolRevision'] = revision
                     # verify pubication status - sample published notebook reference to simtool
                     simToolNotebookMetaData = _getSimToolNotebookMetaData(simToolLocation['notebookPath'])
                     if simToolNotebookMetaData['name'] == simToolLocation['simToolName'] and \
                        simToolNotebookMetaData['revision'] == simToolLocation['simToolRevision'] and \
                        simToolNotebookMetaData['state'] == 'published':
                        simToolLocation['published'] = True
   elif os.path.isfile(simToolName):
      # *.ipynb - must be a local (non-published) notebook
      simToolLocation['notebookPath']    = os.path.realpath(simToolName)
      simToolLocation['simToolName']     = os.path.splitext(os.path.basename(simToolName))[0]
      simToolLocation['simToolRevision'] = None
      simToolLocation['published']       = False
   else:
      if simToolRevision:
         raise FileNotFoundError('Revision "%s" of simtool named "%s" not found' % (simToolRevision,simToolName))
      else:
         raise FileNotFoundError('No simtool named "%s, "' % (simToolName))

   return simToolLocation


def findInstalledSimToolNotebooks(querySimToolName=None,
                                  returnString=True):
   """Find all the revisions of a SimTool.

       Returns:
           Ordered lists of installed and published revisions
   """
   installedSimToolRevisions = {}

   if querySimToolName:
      simToolNames = [querySimToolName]
   else:
      appsPath = os.path.join(os.sep,'apps')
      appsDirs = os.listdir(appsPath)
      simToolNames = []
      for appsDir in appsDirs:
         simToolPath = os.path.join(os.sep,'apps',appsDir)
         if os.path.isdir(simToolPath):
            simToolNames.append(appsDir)
   simToolNames.sort()

   reFiles = re.compile("^r[0-9]+$")
   for simToolName in simToolNames:
      simToolPath = os.path.join(os.sep,'apps',simToolName)
      try:
         dirFiles = os.listdir(simToolPath)
      except:
         pass
      else:
         matchingFiles = filter(reFiles.search,dirFiles)
         simToolRevisions = []
         for matchingFile in matchingFiles:
            try:
               revisionIndex = int(matchingFile[1:])
            except:
               pass
            else:
               simToolRevisions.append(revisionIndex)
         simToolRevisions.sort()
         simToolRevisions = [ 'r%d' % (revision) for revision in simToolRevisions ]

         for simToolRevision in simToolRevisions:
            nbPath = os.path.join(simToolPath,simToolRevision,'simtool',"%s.ipynb" % (simToolName))
            if os.path.exists(nbPath):
               simToolNotebookMetaData = _getSimToolNotebookMetaData(nbPath)
               if   simToolNotebookMetaData['state'] == 'installed':
                  description = _getSimToolDescription(nbPath)
                  if not simToolName in installedSimToolRevisions:
                     installedSimToolRevisions[simToolName] = {}
                  if not 'installed' in installedSimToolRevisions[simToolName]:
                     installedSimToolRevisions[simToolName]['installed'] = {}
                  installedSimToolRevisions[simToolName]['installed'][simToolRevision] = description
               elif simToolNotebookMetaData['state'] == 'published':
                  description = _getSimToolDescription(nbPath)
                  if not simToolName in installedSimToolRevisions:
                     installedSimToolRevisions[simToolName] = {}
                  if not 'published' in installedSimToolRevisions[simToolName]:
                     installedSimToolRevisions[simToolName]['published'] = {}
                  installedSimToolRevisions[simToolName]['published'][simToolRevision] = description

   if returnString:
      installedSimToolRevisions = yaml.dump(installedSimToolRevisions,indent=3)
      installedSimToolRevisions = installedSimToolRevisions.replace("\n\n", "\n  ").strip()

   return installedSimToolRevisions


def searchForSimTool(simToolName,simToolRevision=None):
   """Lookup simtool by name and revision.

      Args:
          simToolName: SimTool name.
          simToolRevision: SimTool revision, typically rNN.

      Returns:
          A simToolLocation dictionary containing
              notebookPath    - the full path name of the simtool notebook,
              simToolName     - the simtool shortname
              simToolRevision - the simtool revision (if installed or published)
              published       - boolean which is True if the notebook is published
   """
   foundIt = True
   if simToolRevision is None:
      notebookPath = os.path.join('simtool',"%s.ipynb" % (simToolName))
      if not os.path.islink(notebookPath):
         notebookPath = os.path.join('..','simtool',"%s.ipynb" % (simToolName))
         if not os.path.islink(notebookPath):
            foundIt = False

      if foundIt:
#        verify link to installed (/apps) version
         if os.path.isfile(notebookPath):
            notebookPath = os.path.realpath(os.path.abspath(notebookPath))
            installedNotebookPattern = os.path.join(os.sep,'apps',simToolName,'(r[0-9]+)','simtool',"%s.ipynb" % (simToolName))
            reInstalledNotebookPattern = re.compile("^%s$" % (installedNotebookPattern))
            match = reInstalledNotebookPattern.match(notebookPath)
            if match:
               simToolLocation = {}
               simToolLocation['notebookPath']    = notebookPath
               simToolLocation['simToolName']     = simToolName
               simToolLocation['simToolRevision'] = match.group(1)
               simToolNotebookMetaData = _getSimToolNotebookMetaData(simToolLocation['notebookPath'])
               if simToolNotebookMetaData['name'] == simToolLocation['simToolName'] and \
                  simToolNotebookMetaData['revision'] == simToolLocation['simToolRevision'] and \
                  simToolNotebookMetaData['state'] == 'published':
                  simToolLocation['published'] = True
               else:
                  simToolLocation['published'] = False
            else:
               foundIt = False
         else:
#           broken link
            foundIt = False

      if not foundIt:
# Not an installed SimTool
         foundIt = True

         notebookPath = os.path.join('simtool',"%s.ipynb" % (simToolName))
         if not os.path.isfile(notebookPath):
            notebookPath = os.path.join('..','simtool',"%s.ipynb" % (simToolName))
            if not os.path.isfile(notebookPath):
               foundIt = False

         if foundIt:
            notebookPath = os.path.realpath(os.path.abspath(notebookPath))
            simToolLocation = {}
            simToolLocation['notebookPath']    = notebookPath
            simToolLocation['simToolName']     = simToolName
            simToolLocation['simToolRevision'] = None
            simToolLocation['published']       = False

      if not foundIt:
         foundIt = True
         try:
            simToolLocation = findSimToolNotebook(simToolName,'current')
         except:
            foundIt = False
         else:
            notebookPath = simToolLocation['notebookPath']
            if notebookPath is None:
               foundIt = False
            else:
               simToolLocation['simToolRevision'] = os.path.basename(os.path.dirname(os.path.dirname(notebookPath)))

      if not foundIt:
         foundIt = True
         try:
            simToolLocation = findSimToolNotebook(simToolName,'dev')
         except:
            foundIt = False
         else:
            notebookPath = simToolLocation['notebookPath']
            if notebookPath is None:
               foundIt = False
            else:
               simToolLocation['simToolRevision'] = os.path.basename(os.path.dirname(os.path.dirname(notebookPath)))
   else:
      try:
         simToolLocation = findSimToolNotebook(simToolName,simToolRevision)
      except:
         foundIt = False
      else:
         if simToolLocation['notebookPath'] is None:
            foundIt = False

   if not foundIt:
      simToolLocation = {}
      simToolLocation['notebookPath']    = None
      simToolLocation['simToolName']     = None
      simToolLocation['simToolRevision'] = None
      simToolLocation['published']       = None

   return simToolLocation


def _find_simTool(simToolName,simToolRevision=None):
    """Lookup simtool by name and revision.

        Returns:
            A tuple containing the full path name of the simtool notebook,
            the tool name, the tool revision (if published) and a boolean which is True if the notebook
            is published
    """
    if   simToolRevision and not simToolName.endswith('.ipynb'):
        simToolNotebook = simToolName + '.ipynb'
        prefix = 'apps'
        if   os.path.exists(os.path.join(prefix,simToolName,simToolRevision,simToolNotebook)):
            return (os.path.join(prefix,simToolName,simToolRevision,simToolNotebook),simToolName,simToolRevision,True)
        elif os.path.exists(os.path.join(simToolName,simToolRevision,simToolNotebook)):
            tool_name = os.path.splitext(os.path.basename(simToolName))[0]
            return (os.path.join(os.path.realpath(simToolName),simToolRevision,simToolNotebook),tool_name,simToolRevision,False)
    elif not simToolName.endswith('.ipynb'):
        simToolNotebook = simToolName + '.ipynb'
        if os.path.exists(os.path.join(simToolName,simToolRevision,simToolNotebook)):
            tool_name = os.path.splitext(os.path.basename(simToolName))[0]
            return (os.path.join(os.path.realpath(simToolName),simToolRevision,simToolNotebook),tool_name,simToolRevision,False)
    elif os.path.isfile(simToolName):
        # must be a local (non-published) notebook
        tool_name = os.path.splitext(os.path.basename(simToolName))[0]
        return (os.path.realpath(simToolName), tool_name, None,False)
    else:
        if simToolRevision:
            raise FileNotFoundError('Revision "%s" of simtool named "%s" not found' % (simToolRevision,simToolName))
        else:
            raise FileNotFoundError('No simtool named "%s, "' % (simToolName))


def _getNotebookeCellYAMLcontent(nb,
                                 yamlTag):
   yamlDict = None
# ignore lines up to and including %%yaml (cell magic)
   yamlContent = None
   yamlLineNumber = -1
   for cell in nb.cells:
      cellSourceLines = cell['source'].split('\n')
      lineNumber = 0
      for cellSourceLine in cellSourceLines:
         if cellSourceLine.startswith("%%%%yaml %s" % (yamlTag)):
            yamlLineNumber = lineNumber
            break
         lineNumber += 1

      if yamlLineNumber >= 0:
         yamlContent = '\n'.join(cellSourceLines[yamlLineNumber+1:])
         break
      
   if yamlContent:
      yamlDict = yaml.load(yamlContent, Loader=yaml.FullLoader)

   return yamlDict


def getNotebookInputs(nb):
   yamlDict = _getNotebookeCellYAMLcontent(nb,"INPUTS")
   if yamlDict:
      return parse(yamlDict)
   else:
      return None


def getSimToolInputs(simToolLocation):
   """Get required SimTool inputs definition.

      Args:
          simToolLocation:  A dictionary containing information on SimTool notebook
              location and status.
      Returns:
          A simtool.Params object defining expected inputs.
   """
   nbPath = simToolLocation['notebookPath']
   nb = load_notebook_node(nbPath)

   return getNotebookInputs(nb)


def _get_inputs_dict(inputs,
                     inputFileRunPrefix=None):
   if type(inputs) == dict:
      return inputs
   inputsDict = {}
   for i in inputs:
      value = inputs[i].serialValue

      checkForFile = False
      try:
         if isinstance(value,basestring):
            checkForFile = True
      except NameError:
         if isinstance(value,str):
            checkForFile = True

      if checkForFile:
         if value.startswith('file://'):
            path = value[7:]
            fileName = os.path.basename(path)
            if inputFileRunPrefix:
               value = 'file://' + os.path.join(inputFileRunPrefix,fileName)
            else:
               value = 'file://' + fileName
      inputsDict[i] = value

   return inputsDict


def _get_inputs_cache_dict(inputs):
   if type(inputs) == dict:
      return inputs
   inputsCacheDict = {}
   for i in inputs:
      value = inputs[i].serialValue

      checkForFile = False
      try:
         if isinstance(value,basestring):
            checkForFile = True
      except NameError:
         if isinstance(value,str):
            checkForFile = True

      if checkForFile:
         if value.startswith('file://'):
            path = value[7:]
            with open(path,'rb') as fp:
               value = fp.read()
      inputsCacheDict[i] = value

   return inputsCacheDict


def _get_inputFiles(inputs):
   if type(inputs) == dict:
      return None

   inputFiles = []
   for i in inputs:
      try:
         if inputs[i].file:
            inputFiles.append(inputs[i].file)
      except:
         pass
   return inputFiles


def getNotebookOutputs(nb):
   yamlDict = _getNotebookeCellYAMLcontent(nb,"OUTPUTS")
   if yamlDict:
      return parse(yamlDict)
   else:
      return None


def getSimToolOutputs(simToolLocation):
   """Get SimTool outputs definition.

      Args:
          simToolLocation:  A dictionary containing information on SimTool notebook
              location and status.
      Returns:
          A simtool.Params object defining expected outputs.
   """
   nbPath = simToolLocation['notebookPath']
   nb = load_notebook_node(nbPath)

   return getNotebookOutputs(nb)





# @package      hubzero-simtool
# @file         run.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import uuid
import copy
import json
import shutil
import tempfile
import stat
from subprocess import call
import traceback
try:
   from hubzero.submit.SubmitCommand import SubmitCommand
except ImportError:
   submitAvailable = False
else:
   submitAvailable = True

import papermill as pm
import yaml
from .db import DB
from .experiment import get_experiment
from .datastore import FileDataStore
from .utils import _get_inputs_dict, _get_extra_files, _get_inputFiles, _get_inputs_cache_dict, getSimToolOutputs


class RunBase:
   """
   Base class for SimTool Run
   """

   DSHANDLER          = FileDataStore  # local files or NFS.  should be config option
   INPUTFILERUNPREFIX = '.notebookInputFiles'
   SIMTOOLRUNPREFIX   = '.simtool'

   def __init__(self,simToolLocation,inputs,runName,cache,
                     remoteAttributes=None,
                     remote=False,trustedExecution=False):
      self.nbName = simToolLocation['simToolName'] + '.ipynb'
      self.inputs = copy.deepcopy(inputs)
      self.input_dict = _get_inputs_dict(self.inputs,inputFileRunPrefix=RunBase.INPUTFILERUNPREFIX)
      self.inputFiles = _get_inputFiles(self.inputs)
      self.outputs = copy.deepcopy(getSimToolOutputs(simToolLocation))

# Create landing area for results
      if runName:
         self.runName = runName
      else:
         self.runName = str(uuid.uuid4()).replace('-','')
      self.outdir = os.path.join(get_experiment(),self.runName)
      os.makedirs(self.outdir)
      self.outname = os.path.join(self.outdir,self.nbName)
      if remote:
         self.remoteSimTool = os.path.join(self.outdir,RunBase.SIMTOOLRUNPREFIX)
         os.makedirs(self.remoteSimTool)
      else:
         self.remoteSimTool = None

      self.cached = False
      self.dstore = None
      if not trustedExecution:
         if cache:
            hashableInputs = _get_inputs_cache_dict(self.inputs)
            self.dstore = RunBase.DSHANDLER(simToolLocation['simToolName'],simToolLocation['simToolRevision'],hashableInputs)
            del hashableInputs
            self.cached = self.dstore.read_cache(self.outdir)

         print("runname   = %s" % (self.runName))
         print("outdir    = %s" % (self.outdir))
         print("cached    = %s" % (self.cached))

      self.inputsPath = None
      self.db = None
      self.savedOutputFiles = None
      self.savedOutputs = None


   @staticmethod
   def __copySimToolTreeAsLinks(sdir,ddir):
      simToolFiles = os.listdir(sdir)
      for simToolFile in simToolFiles:
         simToolPath = os.path.join(sdir,simToolFile)
         if os.path.isdir(simToolPath):
            shutil.copytree(simToolPath,ddir,copy_function=os.symlink)
         else:
            os.symlink(simToolPath,os.path.join(ddir,simToolFile))


   def setupInputFiles(self,simToolLocation,
                            doSimToolFiles=True,keepSimToolNotebook=False,remote=False,
                            doUserInputFiles=True,
                            doSimToolInputFile=True):
      if doSimToolFiles:
         if remote:
            ddir = self.remoteSimTool
         else:
            ddir = self.outdir
         # Prepare output directory by copying any files that the notebook depends on.
         sdir = os.path.abspath(os.path.dirname(simToolLocation['notebookPath']))
         if simToolLocation['published']:
            # We want to allow simtools to be more than just the notebook,
            # so we recursively copy the notebook directory.
            self.__copySimToolTreeAsLinks(sdir,ddir)
            # except the notebook itself
            if not keepSimToolNotebook:
               os.remove(os.path.join(ddir,self.nbName))
         else:
            if keepSimToolNotebook and remote:
               if sdir.startswith('/apps/'):
                  os.symlink(os.path.join(sdir,self.nbName),os.path.join(ddir,self.nbName))
               else:
                  raise ValueError("SimTool notebook must be installed or published to used remotely")
            extraFiles = _get_extra_files(simToolLocation['notebookPath'])
            # print("EXTRA FILES:",extraFiles)
            if   extraFiles == "*":
               self.__copySimToolTreeAsLinks(sdir,ddir)
               if not keepSimToolNotebook:
                  os.remove(os.path.join(ddir,self.nbName))
            elif extraFiles is not None:
               for extraFile in extraFiles:
                  os.symlink(os.path.abspath(os.path.join(sdir,extraFile)),os.path.join(ddir,extraFile))

      if doUserInputFiles:
         inputFileRunPath = os.path.join(self.outdir,RunBase.INPUTFILERUNPREFIX)
         os.makedirs(inputFileRunPath)
         for inputFile in self.inputFiles:
            shutil.copy2(inputFile,inputFileRunPath)

      if doSimToolInputFile:
# Generate inputs file for cache comparison and/or job input
         self.inputsPath = os.path.join(self.outdir,'inputs.yaml')
         with open(self.inputsPath,'w') as fp:
            yaml.dump(self.input_dict,fp)


   def checkTrustedUserCache(self,simToolLocation):
      submitCommand = SubmitCommand()
      submitCommand.setLocal()
      submitCommand.setCommand(os.path.join(os.sep,'apps','bin','ionhelperGetArchivedSimToolResult.sh'))
      submitCommand.setCommandArguments([simToolLocation['simToolName'],
                                         simToolLocation['simToolRevision'],
                                         self.inputsPath,
                                         self.outdir])
      submitCommand.show()
      try:
         result = submitCommand.submit()
      except:
         exitCode = 1
         print(traceback.format_exc())
      else:
         exitCode = result['exitCode']
         if exitCode == 0:
            print("Found cached result")

      self.cached = exitCode == 0


   def doTrustedUserRun(self,simToolLocation,
                             remoteAttributes=None):
      if remoteAttributes:
# pass along remote submit command arguments: venue, walltime, cores, command
         argumentsPath = os.path.join(self.outdir,'remoteArguments.json')
         with open(argumentsPath,'w') as fp:
            json.dump(remoteAttributes,fp)

      submitCommand = SubmitCommand()
      submitCommand.setLocal()
      submitCommand.setCommand(os.path.join(os.sep,'apps','bin','ionhelperRunSimTool.sh'))
      submitCommand.setCommandArguments([simToolLocation['simToolName'],
                                         simToolLocation['simToolRevision'],
                                         self.inputsPath])
      submitCommand.show()
      try:
         result = submitCommand.submit()
      except:
         exitCode = 1
         print(traceback.format_exc())
      else:
         exitCode = result['exitCode']
         if exitCode != 0:
            print("SimTool execution failed")
      self.cached = exitCode == 0


   def retrieveTrustedUserResults(self,simToolLocation):
      if self.cached:
#        Retrieve result from cache
         submitCommand = SubmitCommand()
         submitCommand.setLocal()
         submitCommand.setCommand(os.path.join(os.sep,'apps','bin','ionhelperGetArchivedSimToolResult.sh'))
         submitCommand.setCommandArguments([simToolLocation['simToolName'],
                                            simToolLocation['simToolRevision'],
                                            self.inputsPath,
                                            self.outdir])
         submitCommand.show()
         try:
            result = submitCommand.submit()
         except:
            exitCode = 1
            print(traceback.format_exc())
         else:
            exitCode = result['exitCode']
            if exitCode != 0:
               print("Retrieval of generated cached result failed")
      else:
#        Retrieve error result from ionhelper delivery
         submitCommand = SubmitCommand()
         submitCommand.setLocal()
         submitCommand.setCommand(os.path.join(os.sep,'apps','bin','ionhelperLoadSimToolResult.sh'))
         submitCommand.setCommandArguments([self.outdir])
         submitCommand.show()
         try:
            result = submitCommand.submit()
         except:
            exitCode = 1
            print(traceback.format_exc())
         else:
            exitCode = result['exitCode']
            if exitCode != 0:
               print("Retrieval of failed execution result failed")


   def processOutputs(self,cache,prerunFiles,
                           trustedExecution=False):
      self.db = DB(self.outname,dir=self.outdir)
      if not trustedExecution:
         self.savedOutputs     = self.db.getSavedOutputs()
         self.savedOutputFiles = self.db.getSavedOutputFiles()
#        if len(self.savedOutputFiles) > 0:
#           print("Saved output files: %s" % (self.savedOutputFiles))

         requiredOutputs  = set(self.outputs.keys())
         deliveredOutputs = set(self.savedOutputs)
         missingOutputs = requiredOutputs - deliveredOutputs
         extraOutputs   = deliveredOutputs - requiredOutputs
         if 'simToolSaveErrorOccurred' in extraOutputs:
            extraOutputs.remove('simToolSaveErrorOccurred')
         if 'simToolAllOutputsSaved' in extraOutputs:
            extraOutputs.remove('simToolAllOutputsSaved')
         if len(missingOutputs) > 0:
            print("The following outputs are missing: %s" % (list(missingOutputs)))
         if len(extraOutputs) > 0:
            print("The following additional outputs were returned: %s" % (list(extraOutputs)))

#        simToolSaveErrorOccurred = self.db.getSimToolSaveErrorOccurred()
#        print("simToolSaveErrorOccurred = %d" % (simToolSaveErrorOccurred))
#        simToolAllOutputsSaved = self.db.getSimToolAllOutputsSaved()
#        print("simToolAllOutputsSaved = %d" % (simToolAllOutputsSaved))

         if cache:
            self.dstore.write_cache(self.outdir,prerunFiles,self.savedOutputFiles)


   def getResultSummary(self):
      return self.db.nb.scrap_dataframe


   def read(self, name, display=False, raw=False):
      return self.db.read(name,display,raw)


class LocalRun(RunBase):
   """
   Run a notebook without using submit.
   """

   def __init__(self,simToolLocation,inputs,runName,cache):
      RunBase.__init__(self,simToolLocation,inputs,runName,cache,
                            remoteAttributes=None,
                            remote=False,trustedExecution=False)

      if not self.cached:
         self.setupInputFiles(simToolLocation,
                              doSimToolFiles=True,keepSimToolNotebook=False,remote=False,
                              doUserInputFiles=True,
                              doSimToolInputFile=False)

         prerunFiles = os.listdir(self.outdir)
         prerunFiles.append(self.nbName)

         # FIXME: run in background. wait or check status.
         pm.execute_notebook(simToolLocation['notebookPath'],self.outname,parameters=self.input_dict,cwd=self.outdir)

         self.processOutputs(cache,prerunFiles,trustedExecution=False)
      else:
         self.db = DB(self.outname,dir=self.outdir)


class SubmitLocalRun(RunBase):
   """
   Run a notebook using submit --local.
   """

   def __init__(self,simToolLocation,inputs,runName,cache):
      RunBase.__init__(self,simToolLocation,inputs,runName,cache,
                            remoteAttributes=None,
                            remote=False,trustedExecution=False)

      if not self.cached:
         self.setupInputFiles(simToolLocation,
                              doSimToolFiles=True,keepSimToolNotebook=False,remote=False,
                              doUserInputFiles=True,
                              doSimToolInputFile=True)

         cwd = os.getcwd()
         os.chdir(self.outdir)

         prerunFiles = os.listdir(os.getcwd())
         prerunFiles.append(self.nbName)

         # FIXME: run in background. wait or check status.
         submitCommand = SubmitCommand()
         submitCommand.setLocal()
         submitCommand.setCommand("papermill")
         submitCommand.setCommandArguments(["-f","inputs.yaml",
                                            simToolLocation['notebookPath'],
                                            self.nbName])
         submitCommand.show()
         try:
            result = submitCommand.submit()
         except:
            exitCode = 1
            print(traceback.format_exc())
         else:
            exitCode = result['exitCode']
            if exitCode != 0:
               print("SimTool execution failed")

         os.chdir(cwd)

         self.processOutputs(cache,prerunFiles,trustedExecution=False)
      else:
         self.db = DB(self.outname,dir=self.outdir)


class SubmitRemoteRun(RunBase):
   """
   Run a notebook using submit --venue VENUE -w TIME -n CORES.
   """

   def __init__(self,simToolLocation,inputs,runName,remoteAttributes,cache):
      RunBase.__init__(self,simToolLocation,inputs,runName,cache,
                            remoteAttributes=remoteAttributes,
                            remote=True,trustedExecution=False)

      if not self.cached:
         self.setupInputFiles(simToolLocation,
                              doSimToolFiles=True,keepSimToolNotebook=True,remote=True,
                              doUserInputFiles=True,
                              doSimToolInputFile=True)

         cwd = os.getcwd()
         os.chdir(self.outdir)

         prerunFiles = os.listdir(os.getcwd())
         prerunFiles.append(self.nbName)

         # FIXME: run in background. wait or check status.
         submitCommand = SubmitCommand()
         try:
            submitCommand.setVenue(remoteAttributes['venue'])
         except:
            pass
         try:
            submitCommand.setWallTime(remoteAttributes['wallTime'])
         except:
            pass
         try:
            submitCommand.setNcores(remoteAttributes['nCores'])
         except:
            pass
         submitCommand.setInputFiles([RunBase.SIMTOOLRUNPREFIX,RunBase.INPUTFILERUNPREFIX])
         submitCommand.setCommand(remoteAttributes['command'])
         submitCommand.setCommandArguments(["-s",simToolLocation['simToolName'],
                                            "-i","inputs.yaml"])
         submitCommand.show()
         try:
            result = submitCommand.submit()
         except:
            exitCode = 1
            print(traceback.format_exc())
         else:
            exitCode = result['exitCode']
            if exitCode != 0:
               print("SimTool execution failed")

         shutil.rmtree(self.remoteSimTool,True)

         os.chdir(cwd)

         self.processOutputs(cache,prerunFiles,trustedExecution=False)
      else:
         shutil.rmtree(self.remoteSimTool,True)
         self.db = DB(self.outname,dir=self.outdir)


class TrustedUserLocalRun(RunBase):
   """
   Prepare and run of a notebook as a trusted user.
   """

   def __init__(self,simToolLocation,inputs,runName,cache):
      if simToolLocation['published']:
# Only published simTool can be run with trusted user
         RunBase.__init__(self,simToolLocation,inputs,runName,cache,
                               remoteAttributes=None,
                               remote=False,trustedExecution=True)

         self.setupInputFiles(simToolLocation,
                              doSimToolFiles=False,keepSimToolNotebook=False,remote=False,
                              doUserInputFiles=True,
                              doSimToolInputFile=True)

         self.checkTrustedUserCache(simToolLocation)
         if not self.cached:
            self.doTrustedUserRun(simToolLocation,remoteAttributes=None)
            self.retrieveTrustedUserResults(simToolLocation)

         prerunFiles = None
         self.processOutputs(cache,prerunFiles,trustedExecution=True)
      else:
         print("The simtool %s/%s is not published" % (simToolLocation['simToolName'],simToolLocation['simToolRevision']))


class TrustedUserRemoteRun(RunBase):
   """
   Prepare and run of a notebook with remote execution as a trusted user.
   """

   def __init__(self,simToolLocation,inputs,runName,remoteAttributes,cache):
      if simToolLocation['published']:
# Only published simTool can be run with trusted user
         RunBase.__init__(self,simToolLocation,inputs,runName,cache,
                               remoteAttributes=remoteAttributes,
                               remote=True,trustedExecution=True)

         self.setupInputFiles(simToolLocation,
                              doSimToolFiles=True,keepSimToolNotebook=True,remote=True,
                              doUserInputFiles=True,
                              doSimToolInputFile=True)

         self.checkTrustedUserCache(simToolLocation)
         if not self.cached:
            self.doTrustedUserRun(simToolLocation,remoteAttributes=remoteAttributes)
            shutil.rmtree(self.remoteSimTool,True)
            self.retrieveTrustedUserResults(simToolLocation)
         else:
            shutil.rmtree(self.remoteSimTool,True)

         prerunFiles = None
         self.processOutputs(cache,prerunFiles,trustedExecution=True)
      else:
         print("The simtool %s/%s is not published" % (simToolLocation['simToolName'],simToolLocation['simToolRevision']))


class Run:
   """Runs a SimTool.

       A copy of the SimTool will be created in the subdirectory with the same
       name as the current experiment.  It will be run with the provided inputs.

       If cache is True and the tool is published, the global cache will be used.
       If the tool is not published, and cache is True, a local user cache will be used.

       Args:
           simToolLocation: A dictionary containing information on SimTool notebook
               location and status.
           inputs: A SimTools Params object or a dictionary of key-value pairs.
           runName: An optional name for the run.  A unique name will be generated
               if no name is supplied.
           remoteAttributes: A list of parameters used for submission to offsite
               resource.  In the absense of remoteAttributes the notebook execution
               will occur locally.
           cache: If the SimTool was run with the same inputs previously, return
               the results from the cache.  Otherwise cache the results.  If this
               parameter is False, do neither of these.  The SimTool must be published
               to access the global cache, otherwise each user has a local cache
               that can be accesed.
           venue: [None, noSubmit, local, trustedLocal, remote, trustedRemote] venue selection mode.
               noSubmit - to ignore presense of submit.
               local - to use 'submit --local'.
               trustedLocal - to use 'submit --local' as the trusted user for global cache interaction.
               remote - to use 'submit' to execute job on remote resource.
               trustedRemote - to use 'submit' to execute job on remote resource as the trusted user for global cache interaction.
               None - venue is determined based on the availability of submit and the other arguments.
       Returns:
           A Run object.
       """

   def __new__(cls,simToolLocation,inputs,runName=None,remoteAttributes=None,cache=True,venue=None):
      remoteRunAttributes = copy.deepcopy(remoteAttributes)
      if venue is None and submitAvailable:
         if   remoteRunAttributes:
            if simToolLocation['published'] and cache:
               venue = 'trustedRemote'
            else:
               venue = 'remote'
            if not 'command' in remoteRunAttributes:
               try:
                  nCores = remoteRunAttributes['nCores']
               except:
                  nCores = 1
               if nCores == 1:
                  remoteRunAttributes['command'] = "%s_simtool_serial" % (simToolLocation['simToolName'])
               else:
                  remoteRunAttributes['command'] = "%s_simtool_mpi" % (simToolLocation['simToolName'])
         elif simToolLocation['published'] and cache:
            venue = 'trustedLocal'
         else:
            venue = 'local'

      if simToolLocation['simToolRevision'] is None:
         cache = False

      if   venue == 'local':
         newclass = SubmitLocalRun(simToolLocation,inputs,runName,cache)
      elif venue == 'remote':
         newclass = SubmitRemoteRun(simToolLocation,inputs,runName,remoteRunAttributes,cache)
      elif venue == 'trustedLocal':
         newclass = TrustedUserLocalRun(simToolLocation,inputs,runName,cache)
      elif venue == 'trustedRemote':
         newclass = TrustedUserRemoteRun(simToolLocation,inputs,runName,remoteRunAttributes,cache)
      elif venue == 'noSubmit':
         newclass = LocalRun(simToolLocation,inputs,runName,cache)
      elif venue is None:
         newclass = LocalRun(simToolLocation,inputs,runName,cache)
      else:
         raise ValueError('Bad venue/cache combination')

      return newclass





# @package      hubzero-simtool
# @file         __init__.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
__version__ = '0.3.4'

from .utils import getGetSimToolNameRevisionFromEnvironment, findInstalledSimToolNotebooks, searchForSimTool
from .utils import findInstalledSimToolNotebooks as findSimTools
from .utils import parse, getValidatedInputs
from .utils import findSimToolNotebook, getSimToolInputs, getSimToolOutputs
from .run import Run, DB 
from .experiment import Experiment, set_experiment, get_experiment



# @package      hubzero-simtool
# @file         db.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import copy
import scrapbook as sb
from IPython.display import display as idisplay

from .utils import parse, getNotebookOutputs, Params
from .datastore import FileDataStore
from .encode import JsonEncoder

class DB(object):

    def __init__(self, outputs, dir=None):
        if type(outputs) is str:
            self.nb = sb.read_notebook(outputs)
            self.out = getNotebookOutputs(self.nb)
        else:
            self.out = parse(outputs)
            self.outputsToBeSaved = copy.deepcopy(self.out.keys())
            self.setSimToolSaveErrorOccurred(0)
            self.setSimToolAllOutputsSaved(0)

        self.dir = dir
        if self.dir is None:
            self.dir = os.getcwd()


    def getSimToolSaveErrorOccurred(self):
        simToolSaveErrorOccurred = DB.encoder.decode(self.read('simToolSaveErrorOccurred',display=False,raw=False))
        return simToolSaveErrorOccurred


    def setSimToolSaveErrorOccurred(self, value):
        data = DB.encoder.encode(value)
        sb.glue('simToolSaveErrorOccurred', data)


    def getSimToolAllOutputsSaved(self):
        simToolAllOutputsSaved = DB.encoder.decode(self.read('simToolAllOutputsSaved',display=False,raw=False))
        return simToolAllOutputsSaved


    def setSimToolAllOutputsSaved(self, value):
        data = DB.encoder.encode(value)
        sb.glue('simToolAllOutputsSaved', data)


    @staticmethod
    def _make_ref(filename):
        return 'file://' + filename


    @staticmethod
    def _get_ref(val):
        if type(val) is not str:
            return None
        if val.startswith('file://'):
            return val[7:]
        return None


    def save(self, name, value=None, display=False, file=None, force=False):
        """Save output to the results database.

        Outputs are saved as key-value entries in the notebook metadata.  Large output
        will typically be in files, so the value saved in the notebook will be a filename.

        Args:
            name: Name of the output.
            value: Value. Not required if the output is in a file.
            display: Should the value be displayed as output for the cell?
            file: Name of the file with the value.
            force: Ignore output schema and write value anyway.
        """
        if   name not in self.out and force is False:
            raise ValueError('\"%s\" not in output schema!' % name)
        elif name in self.out:
# do data validation
            simToolObject = copy.deepcopy(self.out[name])
            try:
                if file != None:
                    simToolObject.file = file
                else:
                    simToolObject.value = value
                    value = copy.deepcopy(simToolObject.value)
            except ValueError as e:
                data = DB.encoder.encode(None)
                sb.glue(name, data)
                self.setSimToolSaveErrorOccurred(1)
                raise ValueError("""save output "%s" failed: %s""" % (name,e.args[0]))
            finally:
                del simToolObject

        # FIXME: If output value is oversized, write it to a file
        # and glue a reference.

        data = None
        if file == None:
            path = self._get_ref(value)
            if path:
                if   not os.path.exists(path):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError(f'File "{path}" does not exist.')
                elif not os.path.isfile(path):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError(f'File "{path}" is not a file.')
                if path.startswith("/") or path.startswith(".."):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError('File must be in the local directory.')
                data = self._make_ref(path)

        if not data:
            if file:
                if value:
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise ValueError('Cannot set both "value" and "file" in save()')
                if   not os.path.exists(file):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError(f'File "{file}" does not exist.')
                elif not os.path.isfile(file):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError(f'File "{file}" is not a file.')
                if file.startswith("/") or file.startswith(".."):
                    data = DB.encoder.encode(None)
                    sb.glue(name, data)
                    self.setSimToolSaveErrorOccurred(1)
                    raise FileNotFoundError('File must be in the local directory.')
                data = self._make_ref(file)
            else:
                data = DB.encoder.encode(value)

        sb.glue(name, data)

        if name in self.outputsToBeSaved:
            self.outputsToBeSaved.remove(name)
            if len(self.outputsToBeSaved) == 0:
                self.setSimToolAllOutputsSaved(1)

        if display:
            self._read(name, data, True, False)


    def read(self, name, display=False, raw=False):
        """Read output from the results database.

        Results are saved internally in the notebook.  The internal result
        could be a reference to a local file.

        Args:
            name: Name of the output.
            display: Should the value be displayed as output for the cell?
            raw: If a reference, just return that.
        Returns:
            The saved value.
        """
        value = None
        try:
            data = self.nb.scraps[name].data
        except KeyError:
            print("%s is not available in results" % (name))
        else:
            value = self._read(name, data, display, raw)

        return value


    def _read(self, name, data, display, raw):
        if name in self.out:
            read_type = Params.types[self.out[name].type]
        else:
            read_type = None
        path = self._get_ref(data)
        if path:
            path = os.path.join(self.dir,path)
            if raw:
                return self._make_ref(path)
            val = DB.datastore.readFile(path,read_type)
        else:
            val = DB.datastore.readData(data,read_type)

        if display:
            idisplay(val)
        return val


    def getSavedOutputs(self):
        savedOutputs = self.nb.scraps.keys()
        return savedOutputs


    def getSavedOutputFiles(self):
        savedOutputFiles = []
        savedOutputs = self.nb.scraps.keys()
        for scrap in savedOutputs:
            data = self.nb.scraps[scrap].data
            if data:
                if type(data) is str:
                    if data.startswith('file://'):
                        savedOutputFiles.append(data[7:])
        return savedOutputFiles


DB.encoder   = JsonEncoder()  # encoder to use for serialzation
DB.datastore = FileDataStore  # configure to use shared filesystem as datastore



# @package      hubzero-simtool
# @file         encode.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import jsonpickle

# The purpose of this class is to abstract out 
# the serialization/deserialization of data so
# that we may change the method in the future.

# abstract class (template)
# (Python doesn't need this, but added anyway
# for clarity.)
class Encoder:
    def encode(self, val):
        pass

    def decode(self, val):
        pass

class JsonEncoder(Encoder):
    def encode(self, val):
        return jsonpickle.dumps(val)

    def decode(self, val):
        return jsonpickle.loads(val)
        




#https://stackoverflow.com/questions/26494747/simple-way-to-choose-which-cells-to-run-in-ipython-notebook-during-run-all/43584169
#https://stackoverflow.com/questions/53204167/is-it-possible-to-combine-magics-in-ipython-jupyter

def detectRankAndSize(line):
   import os
   try:
      mpiRankVar = os.environ['MPI_RANK_VAR']
      rank       = int(os.environ[mpiRankVar])
   except:
      rank = 0
   finally:
      os.environ['APP_MPI_RANK'] = str(rank)

   try:
      mpiSizeVar = os.environ['MPI_SIZE_VAR']
      size       = int(os.environ[mpiSizeVar])
   except:
      size = 0
   finally:
      os.environ['APP_MPI_SIZE'] = str(size)

   return rank,size


def skipRank(rank,cell=None):
   '''Skips execution of the current cell if rank equals 0.'''
   if int(rank) > 0:
      return

   get_ipython().run_cell(cell)


def mpiBarrier(line):
   import os.path
   rank,barrierFile = line.strip().split()
   if int(rank) == 0:
      if not os.path.exists(barrierFile):
         fp = open(barrierFile,'w')
         fp.close()
   else:
      try:
         import time
         while not os.path.exists(barrierFile):
            time.sleep(10)
      except:
         pass


def isMPI(line):
   import os
   try:
      size = int(os.environ['APP_MPI_SIZE'])
   except:
      size = 0

   return size > 0


def load_ipython_extension(ipython):
   '''Registers the skipRank magic when the extension loads.'''
   ipython.register_magic_function(skipRank,'cell')
   ipython.register_magic_function(mpiBarrier,'line')
   ipython.register_magic_function(detectRankAndSize,'line')
   ipython.register_magic_function(isMPI,'line')

def unload_ipython_extension(ipython):
   '''Unregisters the skipRank magic when the extension unloads.'''
   del ipython.magics_manager.magics['cell']['skipRank']
   del ipython.magics_manager.magics['line']['mpiBarrier']
   del ipython.magics_manager.magics['line']['detectRankAndSize']
   del ipython.magics_manager.magics['line']['isMPI']





#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['ipython', 'pint', 'numpy', 'papermill', 'jsonpickle', 
                'mendeleev', 'pillow', 'nteract-scrapbook', 'joblib']

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='simtool',
    version='0.3.4',
    description="Functions for creating and running Simulation Tools",
    long_description=readme + '\n\n' + history,
    maintainer="Steven Clark",
    maintainer_email='smclark@sdsc.edu',
    url='https://github.com/hubzero/simtool',
    packages=[
        'simtool',
    ],
    package_dir={'simtool':
                 'simtool'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='simtool',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# @package      hubzero-simtool
# @file         datastore.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import stat
import json
from joblib import Memory
import uuid
import shutil
import warnings
import requests

class FileDataStore:
   """
   A data store implemented on a shared file system.
   """
   USERCACHELOCATIONROOT = os.path.expanduser('~/data')

   def __init__(self,simtoolName,simtoolRevision,inputs,cacheLocationRoot=None):

      if cacheLocationRoot:
         self.cacheLocationRoot = cacheLocationRoot
      else:
         self.cacheLocationRoot = FileDataStore.USERCACHELOCATIONROOT

      self.cachedir    = os.path.join(self.cacheLocationRoot,'.simtool_cache',simtoolName,simtoolRevision)
      self.cachetabdir = os.path.join(self.cacheLocationRoot,'.simtool_cache_table',simtoolName,simtoolRevision)

#     print(simtoolName,simtoolRevision)
#     print(self.cacheLocationRoot)
#     print("cachedir    = %s" % (self.cachedir))
#     print("cachetabdir = %s" % (self.cachetabdir))
      if not os.path.isdir(self.cachedir):
         os.makedirs(self.cachedir)

      memory = Memory(cachedir=self.cachetabdir, verbose=0)

      @memory.cache
      def make_rname(*args):
         # uuid should be unique, but check just in case
         while True:
            fname = str(uuid.uuid4()).replace('-', '')
            if not os.path.isdir(os.path.join(self.cachedir, fname)):
               break
         return fname

#
# suppress this message:
#
# UserWarning: Persisting input arguments took 0.84s to run.
# If this happens often in your code, it can cause performance problems
# (results will be correct in all cases).
# The reason for this is probably some large input arguments for a wrapped
# function (e.g. large strings).
# THIS IS A JOBLIB ISSUE. If you can, kindly provide the joblib's team with an
# example so that they can fix the problem.
#
      with warnings.catch_warnings():
         warnings.simplefilter('ignore')
         self.rdir = os.path.join(self.cachedir, make_rname(inputs))


   @staticmethod
   def __copySimToolTreeAsLinks(sdir,ddir):
      simToolFiles = os.listdir(sdir)
      for simToolFile in simToolFiles:
         simToolPath = os.path.join(sdir,simToolFile)
         if os.path.isdir(simToolPath):
            shutil.copytree(simToolPath,ddir,copy_function=os.symlink)
         else:
            os.symlink(simToolPath,os.path.join(ddir,simToolFile))


   @staticmethod
   def __copySimToolTree(spath,ddir):
      if os.path.isdir(spath):
         sdir = os.path.realpath(os.path.abspath(spath))
         simToolFiles = os.listdir(sdir)
      else:
         sdir = os.path.dirname(os.path.realpath(os.path.abspath(spath)))
         simToolFiles = [os.path.basename(spath)]

      for simToolFile in simToolFiles:
         simToolPath = os.path.join(sdir,simToolFile)
         if os.path.isdir(simToolPath):
            shutil.copytree(simToolPath,os.path.join(ddir,simToolFile))
         else:
            shutil.copy2(simToolPath,os.path.join(ddir,simToolFile))


   def read_cache(self,outdir):
      # reads cache and copies contents to outdir
      if os.path.exists(self.rdir):
#        print("CACHED. Fetching results from %s" % (self.cacheLocationRoot))
         self.__copySimToolTreeAsLinks(self.rdir,outdir)
         return True
      return False


   def write_cache(self,
                   sourcedir,
                   prerunFiles,
                   savedOutputFiles):
      # copy notebook to data store
      os.makedirs(self.rdir)

#     print("write_cache(sourcedir): %s" % (sourcedir))
#     print("write_cache(cachedir): %s" % (self.rdir))
      for prerunFile in prerunFiles:
         self.__copySimToolTree(os.path.join(sourcedir,prerunFile),self.rdir)
      for savedOutputFile in savedOutputFiles:
         self.__copySimToolTree(os.path.join(sourcedir,savedOutputFile),self.rdir)

      for rootDir,dirNames,fileNames in os.walk(self.rdir):
         for fileName in fileNames:
            filePath = os.path.join(rootDir,fileName)
            os.chmod(filePath,os.stat(filePath).st_mode | stat.S_IROTH)
         for dirName in dirNames:
            dirPath = os.path.join(rootDir,dirName)
            os.chmod(dirPath,os.stat(dirPath).st_mode | stat.S_IROTH | stat.S_IXOTH)


   @staticmethod
   def readFile(path, out_type=None):
      """Reads the contents of an artifact file.

      Args:
          path: Path to the artifact
          out_type: The data type
      Returns:
          The contents of the artifact encoded as specified by the
          output type.  So for an Array, this will return a Numpy array,
          for an Image, an IPython Image, etc.
      """
      if out_type is None:
         with open(path, 'rb') as fp:
            res = fp.read()
         return res
      return out_type.read_from_file(path)


   @staticmethod
   def readData(data, out_type=None):
      """Reads the contents of an artifact data.

      Args:
          data: Artifact data
          out_type: The data type
      Returns:
          The contents of the artifact encoded as specified by the
          output type.  So for an Array, this will return a Numpy array,
          for an Image, an IPython Image, etc.
      """
      if out_type is None:
         return data
      return out_type.read_from_data(data)


class WSDataStore:
   """
   A data store implemented as a web service.
   """
   def __init__(self,simtoolName,simtoolRevision,inputs,cacheLocationRoot):

      self.cacheLocationRoot = cacheLocationRoot.rstrip('/') + '/'

      try:
         # Request the signature for the set of inputs
         squidid = requests.get(self.cacheLocationRoot + "squidid",
                                headers = {'Content-Type': 'application/json'},
                                data = json.dumps({'simtoolName':simtoolName,
                                                   'simtoolRevision':simtoolRevision,
                                                   'inputs':inputs}
                                                 )
                               )
         sid = squidid.json()
         # The signature id (squidid) is saved on the rdir variable instead of the path to the directory
         self.rdir = sid['id']
      except Exception as e:
         # If there is any error obtaining the squidid the mode is changed to global. should it be "local"?
         self.rdir = None


   def read_cache(self, outdir):
      # reads cache and copies contents to outdir
      try:
         squidid = self.rdir
         # request the list of files given the squidid
         cachefile = requests.get(self.cacheLocationRoot + "squidlist",
                                  headers = {'Content-Type': 'application/json'},
                                  data = json.dumps({'squidid':squidid})
                                 )
         results = cachefile.json()
         if len(results) == 0:
            return False;
         if not os.path.isdir(outdir):
            os.mkdir(outdir)
         # for each file on the response, download the blob
         for result in results:
            if "_._" in result['name']:
               outputfile = result['name'].split("_._")
               outputname = outputfile[-1]
               outputdir = os.path.join(outdir,os.sep.join(outputfile[:-1]))
               if not os.path.isdir(outputdir):
                  os.makedirs(outputdir)
            else:
               outputdir = outdir
               outputname = result['name']
            # request the file and save on the proper user file directory
            r = requests.get(self.cacheLocationRoot + "files/" + result['id'],
                             headers = {"Cache-Control": "no-cache"},
                             params = {"download": "true"}
                            )
            with open(os.path.join(outputdir,outputname), 'wb') as fp:
               fp.write(r.content)
         return True
      except Exception as e:
         return False


   def write_cache(self,
                   sourcedir,
                   prerunFiles,
                   savedOutputFiles):
      # copy notebook to data store
      cacheFps = []
      try:
         squidid = self.rdir
         cacheFilePaths = []
         # loop prerunFiles, save full paths
         for prerunFile in prerunFiles:
            cacheFilePath = os.path.realpath(os.path.join(sourcedir,prerunFile))
            if os.path.isdir(cacheFilePath):
               for rootPath,dirs,files in os.walk(cacheFilePath):
                  for cacheFile in files:
                     cacheFilePaths.append(os.path.join(rootPath,cacheFile))
            else:
               cacheFilePaths.append(cacheFilePath)
         # loop savedOutputFiles, save full paths
         for savedOutputFile in savedOutputFiles:
            cacheFilePath = os.path.realpath(os.path.join(sourcedir,savedOutputFile))
            if os.path.isdir(cacheFilePath):
               for rootPath,dirs,files in os.walk(cacheFilePath):
                  for cacheFile in files:
                     cacheFilePaths.append(os.path.join(rootPath,cacheFile))
            else:
               cacheFilePaths.append(cacheFilePath)

         # loop all files found and change the path separtor from / to _._, flattening the path.
         cacheFiles = []
         rootPath = os.path.realpath(sourcedir)
         for cacheFilePath in cacheFilePaths:
            relativePath = os.path.relpath(cacheFilePath,rootPath)
            cacheFp = open(relativePath,'rb')
            cacheFps.append(cacheFp)
            if os.path.dirname(relativePath):
               cacheFiles.append(('file',("_._".join(relativePath.split(os.sep)),cacheFp)))
            else:
               cacheFiles.append(('file',cacheFp))

         # Store the files on the server
#        print("squidid: %s" % (squidid))
#        print("files: %s" % (cacheFiles))
         res = requests.put(self.cacheLocationRoot + "squidlist",
                            data = {'squidid':squidid},
                            files = cacheFiles
                           )
         if res.status_code != 200:
            print("res['status_code']: %s" % (res.status_code))
            print("res['reason']: %s" % (res.reason))
            print("res['text']: %s" % (res.text))
      except Exception as e:
#        print("e: %s" % (e))
         raise e;
      finally:
         for cacheFp in cacheFps:
            cacheFp.close()
         del cacheFps


   @staticmethod
   def readFile(path, out_type=None):
      """Reads the contents of an artifact file.

      Args:
          path: Path to the artifact
          out_type: The data type
      Returns:
          The contents of the artifact encoded as specified by the
          output type.  So for an Array, this will return a Numpy array,
          for an Image, an IPython Image, etc.
      """
      if out_type is None:
         with open(path, 'rb') as fp:
            res = fp.read()
         return res
      return out_type.read_from_file(path)


   @staticmethod
   def readData(data, out_type=None):
      """Reads the contents of an artifact data.

      Args:
          data: Artifact data
          out_type: The data type
      Returns:
          The contents of the artifact encoded as specified by the
          output type.  So for an Array, this will return a Numpy array,
          for an Image, an IPython Image, etc.
      """
      if out_type is None:
         return data
      return out_type.read_from_data(data)





# @package      hubzero-simtool
# @file         params.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import sys
import numpy as np
from mendeleev import element
import PIL.Image
from pint import UnitRegistry
from .encode import JsonEncoder


ureg = UnitRegistry()
ureg.autoconvert_offset_to_baseunit = True
Q_ = ureg.Quantity


# A dictionary-like object that can also
# be accessed by attributes.  Note that you
# cannot access attributes by key, only keys
# can be accessed by attributes.
class Params:
    encoder = JsonEncoder()

    def __init__(self, **kwargs):
        self.__members = []

        if hasattr(self, 'property'):
            self['property'] = kwargs.get('property','symbol')

        for k in kwargs:
            if k in ['type','description']:
                self[k] = kwargs[k]

        if hasattr(self, 'units'):
            units = kwargs.get('units')
            if units:
                try:
                    self['units'] = ureg.parse_units(units)
                except:
                    raise ValueError('Unrecognized units: %s' % (units))
        if hasattr(self, 'min'):
            self['min'] = self._getNumericValueFromQuanity(kwargs.get('min'),checkMinMax=False)
        if hasattr(self, 'max'):
            self['max'] = self._getNumericValueFromQuanity(kwargs.get('max'),checkMinMax=False)

        if hasattr(self, 'options'):
            self['options'] = kwargs.get('options')

        for k in kwargs:
            if k in ['value']:
                self[k] = kwargs[k]

        for k in kwargs:
            if k not in ['type','value','description','units','min','max','options','property']:
                print('Parameter type %s does not have %s attribute.' % (self['type'],k),file=sys.stderr)

        for k in kwargs:
            if k in ['units','min','max','options','property']:
                if not hasattr(self, k):
                    print('Parameter type %s does not have %s attribute.' % (self['type'],k),file=sys.stderr)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        setattr(self, key, value)
        if not key in self.__members:
            self.__members.append(key)

    def has_key(self, key):
        return hasattr(self, key)

    def keys(self):
        return self.__members

    def iterkeys(self):
        return self.__members

    def __iter__(self):
        return iter(self.__members)

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self.__members:
            attributeDictionary[attribute] = self[attribute].getAttributeDictionary()
        return attributeDictionary

    def __repr__(self):
        res = []
        for i in self.__members:
            res.append('%s:' % i)
            res.append(self[i].__repr__())
        return '\n'.join(res)

    @staticmethod
    def _make_ref(filename):
        return 'file://' + filename

    @staticmethod
    def read_from_file(path):
        with open(path) as fp:
            return Params.encoder.decode(fp.read())

    @staticmethod
    def read_from_data(data):
        value = None
        if data:
            value = Params.encoder.decode(data)
        return value

# TODO pressure needs treatment similar to temperature
# absolute, gauge, and pressure diff
    def convert(self, newval):
        "unit conversion with special temperature conversion"
        units = self.units
        if units == ureg.degC or units == ureg.kelvin or units == ureg.degF or units == ureg.degR:
            if newval.units == ureg.coulomb:
                # we want temp, so 'C' is degC, not coulombs
                newval = newval.magnitude * ureg.degC
            elif newval.units == ureg.farad:
               # we want temp, so 'F' is degF, not farads
                newval = newval.magnitude * ureg.degF
        elif units == ureg.delta_degC or units == ureg.delta_degF:
            # detect when user means delta temps
            if newval.units == ureg.degC or newval.units == ureg.coulomb:
                newval = newval.magnitude * ureg.delta_degC
            elif newval.units == ureg.degF or units == ureg.farad:
                newval = newval.magnitude * ureg.delta_degF

        return newval.to(units).magnitude

    def _getNumericValueFromQuanity(self,
                                    quantity,
                                    checkMinMax=True):
        numericValue = None
        if quantity is not None:
            if   hasattr(self, 'units') and self.units and type(quantity) == str:
                numericValue = ureg.parse_expression(quantity)
                if hasattr(numericValue, 'units'):
                    numericValue = self.convert(numericValue)
                else:
                    try:
                        numericValue = float(numericValue)
                    except:
                        raise ValueError("%s is not a number" % (str(quantity)))
            elif type(quantity) == str:
                try:
                    numericValue = float(quantity)
                except:
                    raise ValueError("%s is not a number" % (str(quantity)))
            elif type(quantity) == int:
                numericValue = quantity
            elif type(quantity) == float:
                numericValue = quantity
            elif type(quantity) == np.float64:
                numericValue = float(quantity)
            else:
                raise ValueError("%s is not a number (%s)" % (str(quantity),type(quantity)))

            if checkMinMax and numericValue is not None:
                if self.min is not None and numericValue < self.min:
                    raise ValueError("Minimum value is %g" % (self.min))
                if self.max is not None and numericValue > self.max:
                    raise ValueError("Maximum value is %g" % (self.max))

        return numericValue

    def _getNumericValueForAllQuanities(self,
                                        quantities,
                                        checkMinMax=True):
        for quantity in quantities:
            if isinstance(quantity,list):
                yield list(self._getNumericValueForAllQuanities(quantity,checkMinMax=checkMinMax))
            else:
                yield self._getNumericValueFromQuanity(quantity,checkMinMax=checkMinMax)


class Boolean(Params):
    def __init__(self, **kwargs):
        self._value = None
        super(Boolean, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        if newval is not None:
            if type(newval) != bool:
                raise ValueError("%s is not type bool" % (str(newval)))
        self._value = newval

    @property
    def serialValue(self):
        return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % (self.value)
        return res


class Integer(Params):
    def __init__(self, **kwargs):
        self._value = None
        self.min = None
        self.max = None
        super(Integer, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        if newval is not None:
            if type(newval) == str:
                try:
                    newval = int(newval)
                except:
                    raise ValueError("%s is not an integer" % (newval))
            if self.min is not None and newval < self.min:
                raise ValueError("Minimum value is %d" % (self.min))
            if self.max is not None and newval > self.max:
                raise ValueError("Maximum value is %d" % (self.max))
        self._value = newval

    @property
    def serialValue(self):
        return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Text(Params):
    def __init__(self, **kwargs):
        self._value = None
        self._file  = None
        super(Text, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if newval is not None:
            try:
                if isinstance(newval,basestring):
                    self._value = newval
                else:
                    raise ValueError("%s is not a string" % (str(newval)))
            except NameError:
                if isinstance(newval,str):
                    self._value = newval
                else:
                    raise ValueError("%s is not a string" % (str(newval)))

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, newval):
        self._file = None
        if newval:
            try:
                if os.path.exists(newval):
                    if os.path.isfile(newval):
                        self._file = newval
            except:
                pass

    @property
    def serialValue(self):
        if self._file:
            return self._make_ref(self._file)
        else:
            return self._value

    @staticmethod
    def read_from_file(path):
        with open(path) as fp:
            return fp.read()

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Choice(Params):
    def __init__(self, **kwargs):
        self._value   = None
        self.options = None
        super(Choice, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if self.options is not None:
            try:
                if isinstance(newval,basestring):
                    if newval in self.options:
                        self._value = newval
                    else:
                        raise ValueError("%s is not a valid option" % (str(newval)))
                else:
                    raise ValueError("%s is not a string" % (str(newval)))
            except NameError:
                if isinstance(newval,str):
                    if newval in self.options:
                        self._value = newval
                    else:
                        raise ValueError("%s is not a valid option" % (str(newval)))
                else:
                    raise ValueError("%s is not a string" % (str(newval)))

    @property
    def serialValue(self):
        return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class List(Params):
    def __init__(self, **kwargs):
        self._value = None
        self._file  = None
        super(List, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if newval is not None:
            if   isinstance(newval,list):
                self._value = newval
            elif isinstance(newval,tuple):
                self._value = newval
            else:
                raise ValueError("%s is not a list" % (str(newval)))

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, newval):
        self._file = None
        if newval:
            try:
                if os.path.exists(newval):
                    if os.path.isfile(newval):
                        self._file = newval
            except:
                pass

    @property
    def serialValue(self):
        if self._file:
            return self._make_ref(self._file)
        else:
            return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Dict(Params):
    def __init__(self, **kwargs):
        self._value = None
        self._file  = None
        super(Dict, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if newval is not None:
            if isinstance(newval,dict):
                self._value = newval
            else:
                raise ValueError("%s is not a dictionary" % (str(newval)))

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, newval):
        self._file = None
        if newval:
            try:
                if os.path.exists(newval):
                    if os.path.isfile(newval):
                        self._file = newval
            except:
                pass

    @property
    def serialValue(self):
        if self._file:
            return self._make_ref(self._file)
        else:
            return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Array(Params):
    def __init__(self, **kwargs):
        self._value = None
        self._file  = None
        self.units = None
        self.min = None
        self.max = None
        super(Array, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if newval is not None:
            if   type(newval) is np.ndarray:
                # papermill expects inputs to be json-encodeable by nbformat.
                # This is OK for typical input arrays, but if we ever need
                # to support really large arrays we will need to write a
                # custom papermill engine.
                if self.min is not None and newval.min() < self.min:
                    raise ValueError("Minimum value is %g" % self.min)
                if self.max is not None and newval.max() > self.max:
                    raise ValueError("Maximum value is %g" % self.max)
                self._value = newval.tolist()
            elif type(newval) is list:
                self._value = list(self._getNumericValueForAllQuanities(newval))
            else:
                raise ValueError("%s is not an array" % (str(newval)))

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, newval):
        self._file = None
        if newval:
            try:
                if os.path.exists(newval):
                    if os.path.isfile(newval):
                        self._file = newval
            except:
                pass

    @property
    def serialValue(self):
        if self._file:
            return self._make_ref(self._file)
        else:
            return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if   attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            elif attribute == 'units':
                attributeDictionary[attribute] = "%s" % (self[attribute])
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Number(Params):
    def __init__(self, **kwargs):
        self._value = None
        self.units = None
        self.min = None
        self.max = None
        super(Number, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = self._getNumericValueFromQuanity(newval)

    @property
    def serialValue(self):
        return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if   attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            elif attribute == 'units':
                attributeDictionary[attribute] = "%s" % (self[attribute])
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res


class Image(Params):
    def __init__(self, **kwargs):
        self._value       = None
        self._file        = None
        self._imageFormat = None
        super(Image, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        self._value = None
        if newval is not None:
            if newval:
                try:
                    self._imageFormat = newval.format
                except:
                    if   type(newval) is list:
                        self._value = newval
                    elif type(newval) is np.ndarray:
                        self._value = newval.tolist()
#                   else:
#                       print("Image.value: format failed")
#                       print("type(newval): %s" % (type(newval)))
                else:
                    self._value = np.array(newval).tolist()

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, newval):
        self._file = None
        if newval:
            try:
                if os.path.exists(newval):
                    if os.path.isfile(newval):
                        try:
                            fileImage = self.read_from_file(newval)
                            self._imageFormat = fileImage.format
                        except:
                            raise ValueError("%s is not an image file" % (newval))
                        else:
                            self._file = newval
            except:
                raise ValueError("%s is not an image file" % (newval))

    @property
    def serialValue(self):
        if self._file:
            return self._make_ref(self._file)
        else:
            return self._value

    @property
    def imageFormat(self):
        return self._imageFormat

    @staticmethod
    def read_from_file(path):
        return PIL.Image.open(path)

    @staticmethod
    def read_from_data(data):
        value = None
        if data:
            ordinaryData = Params.encoder.decode(data)
            if ordinaryData:
                npData = np.array(ordinaryData,dtype='uint8')
                value = PIL.Image.fromarray(npData)
        return value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                value = self.serialValue
                if value.startswith('file://'):
                    attributeDictionary[attribute] = value
                else:
                    attributeDictionary[attribute] = '<image>'
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: <image>\n'
        return res


class Element(Params):
    def __init__(self, **kwargs):
        self._value = None
        self.property = None
        super(Element, self).__init__(**kwargs)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newval):
        if type(newval) is not str:
            self._value = newval
            return

        self._e = element(newval.title())
        try:
            self._value = self._e.__dict__[self.property]
        except KeyError:
            print("Error: unknown property:", self.property)
            print("Valid properties are")
            print(list(sorted(self._e.__dict__.keys())))
            raise ValueError("%s is not an element" % (str(newval)))

    @property
    def serialValue(self):
        return self._value

    def getAttributeDictionary(self):
        attributeDictionary = {}
        for attribute in self:
            if attribute == 'value':
                attributeDictionary[attribute] = self.serialValue
            else:
                attributeDictionary[attribute] = self[attribute]
        return attributeDictionary

    def __repr__(self):
        # print all attributes with value last
        res = ''
        for i in self:
            if i != 'value':
                res += '    %s: %s\n' % (i, self[i])
        if self.value is not None:
            res += '    value: %s\n' % self.value
        return res

# register param types
# Dictionary that maps strings to class names.
Params.types = {
    'Boolean': Boolean,
    'Integer': Integer,
    'Text': Text,
    'Choice': Choice,
    'List': List,
    'Dict': Dict,
    'Array': Array,
    'Number': Number,
    'Image': Image,
    'Element': Element
}




# @package      hubzero-simtool
# @file         experiment.py
# @copyright    Copyright (c) 2019-2021 The Regents of the University of California.
# @license      http://opensource.org/licenses/MIT MIT
# @trademark    HUBzero is a registered trademark of The Regents of the University of California.
#
import os
import shutil

class Exp:
    """Experiment class without context manager"""
    def __init__(self, name, append=True):
        if name.startswith('.'):
            raise ValueError('Invalid experiment name')
        if append is False and os.path.exists(name):
            shutil.rmtree(name)
        if not os.path.exists(name):
            os.makedirs(name)
        self.name = name

    def __str__(self):
        return self.name

class Experiment(Exp):
    """Content manager for Experiments

    Create a subdirectory named {name} in which to place new runs.
    If *append* is False, any existing subdirectory with the same
    name is removed.

    You can also use set_experiment() to set the current experiment.

        with Experiment('my_test', append=False):
            Run(nb, inputs1)
            Run(nb, inputs2)

    Attributes:
        name: The experiment name.  When this context is entered
            it will create a subdirectory named {name}.
        append: If True, adds runs to any existing experiment of the
            same name.  If False, removes previous runs.

    """
    _experiments = []  # default name
    active = None

    def __init__(self, name, append=True):
        self.name = name
        Exp.__init__(self, name, append)

    def __enter__(self):
        Experiment._experiments.append(self)
        Experiment.active = self
    def __exit__(self, type, value, traceback):
        Experiment._experiments.pop()
        if Experiment._experiments:
            Experiment.active = Experiment._experiments[-1]
        else:
            Experiment.active = None


def set_experiment(name, append=True):
    """Create a new experiment.

    Create a subdirectory named {name} in which to place new runs.
    If *append* is False, any existing subdirectory with the same
    name is removed.  The default experiment name is 'RUNS'.

    You can also use the Experiment content manager to set the current
    experiment.

    Args:
        name: The name of the experiment (and the subdirectory).
        append: If True, adds runs to any existing experiment of the
            same name.  If False, removes previous runs.
    """
    exp = Exp(name, append)
    Experiment._experiments = [exp]
    Experiment.active = exp

# set_experiment('RUNS')  # set the default name

def get_experiment():
    """Get the current experiment name.

    Returns:
        The current experiment name.  This is the name of directory
        where new runs will be placed.
    """
    if Experiment.active is None:
        return 'RUNS'
    return Experiment.active.name
